<?php

namespace Smarty\Template;

use Smarty\Template;

/**
 * Represents a compiled version of a template or config file.
 * @author     Rodney Rehm
 */
class Compiled extends GeneratedPhpFile {

	/**
	 * nocache hash
	 *
	 * @var string|null
	 */
	public $nocache_hash = null;

	/**
	 * Included sub templates
	 * - index name
	 * - value use count
	 *
	 * @var int[]
	 */
	public $includes = [];

	/**
	 * get a Compiled Object of this source
	 *
	 * @param Template $_template template object
	 *
	 * @return Compiled compiled object
	 */
	public static function load($_template) {
		$compiled = new Compiled();
		if ($_template->getSource()->handler->supportsCompiledTemplates()) {
			$compiled->populateCompiledFilepath($_template);
		}
		return $compiled;
	}

	/**
	 * populate Compiled Object with compiled filepath
	 *
	 * @param Template $_template template object
	 **/
	private function populateCompiledFilepath(Template $_template) {
		$source = $_template->getSource();
		$smarty = $_template->getSmarty();
		$this->filepath = $smarty->getCompileDir();
		if (isset($_template->compile_id)) {
			$this->filepath .= preg_replace('![^\w]+!', '_', $_template->compile_id) .
				($smarty->use_sub_dirs ? DIRECTORY_SEPARATOR : '^');
		}
		// if use_sub_dirs, break file into directories
		if ($smarty->use_sub_dirs) {
			$this->filepath .= $source->uid[0] . $source->uid[1] . DIRECTORY_SEPARATOR . $source->uid[2] .
				$source->uid[3] . DIRECTORY_SEPARATOR . $source->uid[4] . $source->uid[5] .
				DIRECTORY_SEPARATOR;
		}
		$this->filepath .= $source->uid . '_';
		if ($source->isConfig) {
			$this->filepath .= (int)$smarty->config_read_hidden + (int)$smarty->config_booleanize * 2 +
				(int)$smarty->config_overwrite * 4;
		} else {
			$this->filepath .= (int)$smarty->merge_compiled_includes + (int)$smarty->escape_html * 2 +
				(($smarty->merge_compiled_includes && $source->type === 'extends') ? 4 : 0);
		}
		$this->filepath .= '.' . $source->type;
		$basename = $source->handler->getBasename($source);
		if (!empty($basename)) {
			$this->filepath .= '.' . $basename;
		}
		if ($_template->caching) {
			$this->filepath .= '.cache';
		}
		$this->filepath .= '.php';
		$this->timestamp = $this->exists = is_file($this->filepath);
		if ($this->exists) {
			$this->timestamp = filemtime($this->filepath);
		}
	}

	/**
	 * render compiled template code
	 *
	 * @param Template $_template
	 *
	 * @return string
	 * @throws \Smarty\Exception
	 */
	public function render(Template $_template) {
		// checks if template exists
		if (!$_template->getSource()->exists) {
			$type = $_template->getSource()->isConfig ? 'config' : 'template';
			throw new \Smarty\Exception("Unable to load {$type} '{$_template->getSource()->type}:{$_template->getSource()->name}'");
		}
		if ($_template->getSmarty()->debugging) {
			$_template->getSmarty()->getDebug()->start_render($_template);
		}
		if (!$this->processed) {
			$this->process($_template);
		}

		$_template->getCached()->file_dependency =
			array_merge($_template->getCached()->file_dependency, $this->file_dependency);

		$_template->getRenderedTemplateCode($this->unifunc);

		if ($_template->caching && $this->getNocacheCode()) {
			$_template->getCached()->hashes[$this->nocache_hash] = true;
		}
		if ($_template->getSmarty()->debugging) {
			$_template->getSmarty()->getDebug()->end_render($_template);
		}
	}

	/**
	 * load compiled template or compile from source
	 *
	 * @param Template $_smarty_tpl do not change variable name, is used by compiled template
	 *
	 * @throws Exception
	 */
	private function process(Template $_smarty_tpl) {
		$source = $_smarty_tpl->getSource();
		$smarty = $_smarty_tpl->getSmarty();
		if ($source->handler->recompiled) {
			$source->handler->process($_smarty_tpl);
		} else {
			if (!$this->exists || $smarty->force_compile
				|| ($_smarty_tpl->compile_check && $source->getTimeStamp() > $this->getTimeStamp())
			) {
				$this->compileTemplateSource($_smarty_tpl);
				$compileCheck = $_smarty_tpl->compile_check;
				$_smarty_tpl->compile_check = \Smarty\Smarty::COMPILECHECK_OFF;
				$this->loadCompiledTemplate($_smarty_tpl);
				$_smarty_tpl->compile_check = $compileCheck;
			} else {
				$_smarty_tpl->mustCompile = true;
				@include $this->filepath;
				if ($_smarty_tpl->mustCompile) {
					$this->compileTemplateSource($_smarty_tpl);
					$compileCheck = $_smarty_tpl->compile_check;
					$_smarty_tpl->compile_check = \Smarty\Smarty::COMPILECHECK_OFF;
					$this->loadCompiledTemplate($_smarty_tpl);
					$_smarty_tpl->compile_check = $compileCheck;
				}
			}
			$this->processed = true;
		}
	}

	/**
	 * compile template from source
	 *
	 * @param Template $_template
	 *
	 * @throws Exception
	 */
	public function compileTemplateSource(Template $_template) {
		$this->file_dependency = [];
		$this->includes = [];
		$this->nocache_hash = null;
		$this->unifunc = null;
		// compile locking
		if ($saved_timestamp = (!$_template->getSource()->handler->recompiled && is_file($this->filepath))) {
			$saved_timestamp = $this->getTimeStamp();
			touch($this->filepath);
		}
		// compile locking
		try {
			// call compiler
			$this->write($_template, $_template->getCompiler()->compileTemplate($_template));
		} catch (\Exception $e) {
			// restore old timestamp in case of error
			if ($saved_timestamp && is_file($this->filepath)) {
				touch($this->filepath, $saved_timestamp);
			}
			throw $e;
		}
	}

	/**
	 * Write compiled code by handler
	 *
	 * @param Template $_template template object
	 * @param string $code compiled code
	 *
	 * @return bool success
	 * @throws \Smarty\Exception
	 */
	public function write(Template $_template, $code) {
		if (!$_template->getSource()->handler->recompiled) {
			if ($_template->getSmarty()->writeFile($this->filepath, $code) === true) {
				$this->timestamp = $this->exists = is_file($this->filepath);
				if ($this->exists) {
					$this->timestamp = filemtime($this->filepath);
					return true;
				}
			}
			return false;
		}
		return true;
	}

	/**
	 * Read compiled content from handler
	 *
	 * @param Template $_template template object
	 *
	 * @return string content
	 */
	public function read(Template $_template) {
		if (!$_template->getSource()->handler->recompiled) {
			return file_get_contents($this->filepath);
		}
		return false;
	}

	/**
	 * Load fresh compiled template by including the PHP file
	 * HHVM requires a workaround because of a PHP incompatibility
	 *
	 * @param \Smarty\Template $_smarty_tpl do not change variable name, is used by compiled template
	 */
	private function loadCompiledTemplate(Template $_smarty_tpl) {
		if (function_exists('opcache_invalidate')
			&& (!function_exists('ini_get') || strlen(ini_get("opcache.restrict_api")) < 1)
		) {
			opcache_invalidate($this->filepath, true);
		} elseif (function_exists('apc_compile_file')) {
			apc_compile_file($this->filepath);
		}
		if (defined('HHVM_VERSION')) {
			eval('?>' . file_get_contents($this->filepath));
		} else {
			include $this->filepath;
		}
	}

}
